#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "tacoma.h"
#include "meet.h"

#ifndef PAGE_SIZE
# define PAGE_SIZE (1 << 12)
#endif /* PAGE_SIZE */

int ag_c( briefcase bc) {
  cabinet com_ccab,sys_ccab;
  char **af_ptr, **cf_ptr, **f_ptr, **arg_p;
  char **argv;
  char *s_ptr, *e_ptr;
  char tmps[256];
  char com_name[256];
  char cab_path[256];
  int t,rc,pfd[2],pos,size;
  pid_t cpid;

  sprintf(cab_path,"%s/cabinets/C_files",getenv(TAC_ENV_PATH));
  /* 
     Create a compile cabinet. This cabinet is not cleaned, yet...
     */ 
  sprintf(com_name,"cc.%s",*(folder_look(bc,TAC_AGENTID)));
  if (!(com_ccab=cabinet_open(com_name, TC_CREAT))) {
    folder_store(bc, TAC_STATUS, 
		 "Internal failure: unable to create compile cabinet.",
		 NULL);
    return(-1);
  }

  /* 
     Get system specific compiler line 
     */ 
  af_ptr=NULL; 
  if (!((sys_ccab=cabinet_open("C_files", NULL)) &&
      (af_ptr=folder_fetch(sys_ccab,"COMPILE")) && 
      *af_ptr)) {
    folder_store(bc, TAC_STATUS, 
		 "Internal failure: no COMPILE folder configured.",
		 NULL);
    free_copy(af_ptr);      /* Accepts NULL */
    cabinet_close(sys_ccab); /* Accepts NULL */
    cabinet_close(com_ccab); /* Accepts NULL */
    return(-1);
  }

  /* 
     Count the number of *.c files to be compiled
     */
  for (t=0, cf_ptr=folder_look(bc,"CCODE"); *cf_ptr; cf_ptr++) {
    if (f_ptr=folder_look(bc,*cf_ptr)) {
      folder_storeA(com_ccab,*cf_ptr,f_ptr);
      if (!strncmp(&(*cf_ptr)[strlen(*cf_ptr)-2],".c",2)) {
	t++;
      }
    }
  }

  if (!t) {
    folder_store(bc, TAC_STATUS, 
		 "No code found, cannot compile agent.",
		 NULL);
    free_copy(f_ptr);      /* Accepts NULL */
    cabinet_close(sys_ccab); /* Accepts NULL */
    cabinet_close(com_ccab); /* Accepts NULL */
    return(-1);
  }    
  /* 
     Count the number of compiler arguments
     */
  for (arg_p=af_ptr; *arg_p; arg_p++, t++);
  argv=malloc(sizeof(char *)*(t+2)); /*Add NULL & DEF  */

  /* 
     Parse each compiler argument. Works a bit like sprintf with these flags:
     %p - replace with PATH to libraries and includes
     %t - replace with target name
     %c - inserts list of c-files to compile
     */

  for (t=0,arg_p=af_ptr; *arg_p; arg_p++) {
    if (s_ptr=strchr(*arg_p,'%')) {
      switch (s_ptr[1]) {
      case 'c':
	strcpy(tmps,cab_path);
	strcat(tmps,"/c_stub.c");
	argv[t++]=strdup(tmps);
	while (*(e_ptr=folder_rotate(bc,"CCODE",TAC_LEFT,1))) {
	  if (cf_ptr=folder_look(bc,e_ptr)) {
	    folder_storeA(com_ccab,e_ptr,cf_ptr);
	    if (!strncmp(&(e_ptr[strlen(e_ptr)-2]),".c",2)) {
	      argv[t++]=e_ptr;
	    } 
	    else {
	      free(e_ptr);
	    }
	  }
	}
	break;
      case 'l':
      case 'i':
	*s_ptr='\0';
	sprintf(tmps,"%s%s",*arg_p,cab_path);
	argv[t++]=strdup(tmps);
	break;
      case 't':
	*s_ptr='\0';
	sprintf(tmps,"%s%s",*arg_p,*(folder_look(bc,TAC_AGENTID)));
	argv[t++]=strdup(tmps);
	break;
      default:
	argv[t++]=strdup(s_ptr);
	break;
      }
    } 
    else {
      argv[t++]=strdup(*arg_p);
    }
  }    
  free_copy(af_ptr);

  /*
    Find entry point (agent main). If no main is give, call `Main'
    */
  strcpy(tmps,"-DMAIN=");
  if ((e_ptr=folder_rotate(bc,"CMAIN",TAC_LEFT,1)) && (*e_ptr)) {
    strcat(tmps,e_ptr);
    free(e_ptr);
  } 
  else {
    strcat(tmps,"Main");
  }
  argv[t++]=strdup(tmps);
  argv[t++]=NULL;

  /*
    Try to compile the agent
    */
  sprintf(cab_path,"%s/cabinets/%s/",getenv(TAC_ENV_PATH),com_name);
  chdir(cab_path);
  /*
    Setup a pipe from which we can read the process of $CC
    */
  if ( pipe(pfd) == -1 ) {
    folder_store(bc, TAC_STATUS, 
		 "Internal failure: unable to pipe().\n",
		 NULL);
    return(-1);
  }
  switch (cpid=vfork()) {
  case -1:
    folder_store(bc, TAC_STATUS, 
		 "Internal failure: unable to fork().\n",
		 NULL);
    return(-1);
    break;
  case 0:
    close(2);
    dup(pfd[1]); 
    close(pfd[0]);
    execvp(argv[0],argv);
    fprintf(stderr,"Internal failure: unable to execvp().\n");
    close(pfd[1]);
    _exit(1);
    break;
  default:
      close(pfd[1]);
      s_ptr=NULL;
      if (s_ptr=malloc(PAGE_SIZE)) 
	  for ( size=PAGE_SIZE, pos=0; 
	       (t=read(pfd[0], &s_ptr[pos], size-pos))>0;) 
	      if ( (pos+=t) == size )
		  if ( (s_ptr = realloc(s_ptr, (size+=PAGE_SIZE))) == NULL )
		      break;
      wait(&rc);
      close(pfd[0]);
      free_copy(argv);
      break;
  }
  cabinet_close(sys_ccab);
  if (!WIFEXITED(rc) || WEXITSTATUS(rc)!=0) {
      if (s_ptr) 
	  folder_store(bc,TAC_STATUS,s_ptr,NULL);
      else 
	  folder_store(bc,TAC_STATUS,
		       "Compile failure, could not allocate memory for status."
		       ,NULL);
      return(-WEXITSTATUS(rc));
  }
  /* 
     Agent compiled without errors, now try to meet it
     */
  strcat(cab_path,*(folder_look(bc,TAC_AGENTID)));
  rc=meet(cab_path,bc,TM_LOCL|TM_SYNC|TM_PATH);
  chdir(getenv(TAC_ENV_PATH));

  /* 
     Here's a good place the clean the compile cabinet with 
     cabinet_discard(), but I'll leave it in this version

     cabinet_discard(com_ccab);
     */

  cabinet_close(com_ccab);
  

  return(rc);
}
