.TH tac_firewall 1 "TACOMA 1.1.3" "Firewall"
.SH NAME
tac_firewall \- receive incoming briefcases at a site
.SH SYNOPSIS
.B tac_firewall
[\fI-r\fR]
.SH DESCRIPTION
\fBTac_firewall\fR creates a new bridge head for incoming agents
at the current site. It is the single most important command in
the TACOMA system. Its task is to receive incoming briefcases and
execute them locally as agents.

When started \fBtac_firewall\fR will output some diagnostic information and
block waiting for incoming briefcases. It receives these
briefcases through a RPC call which is performed by 
.IR meet (n).

All briefcases that arrive as part of an asynchronous meeting are logged to
disk (unless the briefcase specifies no logging, see
.IR meet (n)). 
Incoming briefcases are stored into the inlog cabinet (cabinets/inlog/).

.SH OPTIONS
.TP
.B -r
This switch is only usable if \fBtac_firewall\fR initially runs as a root
process. \fBTac_firewall\fR will set the file system root to $TACOMAPATH,
see
.IR chroot (2).
The advantage of this is that agents executing code are unable to access
anything outside the subtree of the file system rooted at $TACOMAPATH.
However, this currently also affects all auxiliary programs used by the
TACOMA system. This means that parts of the original file system will have to
be duplicated under the new root. The complete list of files and
directories which need to be accessible under the new root depend strongly 
upon the configuration of TACOMA system. In the future the re-rooting of the
file system may be moved further down the execution path of an agent in
order to lessen its effect on the rest of the TACOMA system. The real user 
id is reset before any system agents are activated.

Notice,
although this functionality gives the file system some protection against 
hostile agents, agents are still able to flood the memory and process
tables of the machine by issuing repeated calls to
.IR malloc (3)
and 
.IR fork (2).

.SH ENVIRONMENT
.TP
TACOMAPATH
This should point to the directory where TACOMA is installed, e.g. the
directory which contains the
cabinets, sysagents, and bin subdirectories. Leave out the trailing slash.
.TP
TACOMAPORT
This is the portnumber on which the firewall will listen for incoming
briefcases.
.SH FILES
All the files described below are multi part cabinet folders. This means
that in the current version of TACOMA they should be accompanied by a file
with the same name, but prefixed by '*#'. These folders are read only once
by the firewall (when it initializes). Changes done to these files will not
be noticed by an already running firewall. Currently a firewall needs to 
be restarted in order for it to notice any changes to these files.
.TP
$TACOMAPATH/cabinets/system/StartupAgents
This folder should contain a list of agents (filenames) which the firewall
will activate when it starts up. No briefcases are delivered to these
agents, they are simply executed silently in the background.
.TP
$TACOMAPATH/cabinets/system/AcceptFrom
This folder is a list of principals from which the firewall will accept
briefcases once authenticated. The exact format of a principal identity 
depend on the authentication scheme used by the incoming briefcase and
.IR ag_crypt (n).
.TP
$TACOMAPATH/cabinets/system/AutoAcceptHost
This folder should contain a list of hosts from which the firewall will accept
briefcases 
.B without 
question. This means that any connection and request to
place a briefcase from a host in this list will be accepted without the
usual authentication scheme. Avoiding the authentication step increases
performance, but needless to say, has serious security implications.
Briefcases accepted in this way are completely anonymous.

.SH SEE ALSO
.IR meet (n)

.SH AUTHOR
Nils Peter Sudmann (nilss@stud.cs.uit.no) and Magne Rasmussen (magner@stud.cs.uit.no), University of Tromsoe. 
Concept by Dag Johansen (dag@cs.uit.no), Robbert van Renesse
(rvr@cs.cornell.edu) and Fred B. Schneider (fbs@cs.cornell.edu).



