'\"     # CS - begin code section
.de CS
.RS 4
\fC
.ps -1
.nf
.sp
..
'\"     # CE - end code section
.de CE
.fi
\fR
.ps +1
.RE
.sp .5
..
.TH TACOMA 3 "TACOMA 1.1.3" "Introduction"
.SH NAME
TACOMA \- execution environment for agents
.SH DESCRIPTION
TACOMA is a joint project between the University of Tromsoe and the Cornell
University. We focus on how to provide OS support for agents that can
roam the internet.

The TACOMA system uses these basic terms:
.TP 10
\fBAgent\fR
- an entity which by it's own initiative moves around a network
.TP
\fBBriefcase\fR
- the set of data needed by an agent on the move is collected
into a briefcase.
.TP
\fBCabinet\fR
- a site bound collection of persistent data.
.TP
\fBFolder\fR
- a subset of the data in a briefcase or a cabinet,
identified by a folder name. Folders may be further divided into
elements of data organized in a list structure.

.SH ""
Briefcases will be discussed first, followed by folders
and agents. Cabinets will not be discussed here. For further
information see 
.IR cabinet_create (3C) 
and 
.IR folder_fetch (3C).

.SH BRIEFCASES
A briefcase contains all the data necessary to enable an
agent to move and execute at different sites. Note that this
does not include data which are bound to a site and which may be used
by the agent when it executes at that site. However, it includes
any data picked up by the agent during its lifetime.
Agents may add (pick up) data to or subtract (drop) data from
a briefcase during its lifetime. Briefcases must be
created implicitly before being used, see 
.IR bc_create (3C).

.SH FOLDERS
The data in a briefcase is organized into folders. Each
folder is identified by a unique name, usually chosen
by the application programmer. However, most system agents
have reserved certain folder names and assigned a certain meaning and
structure to these folders. The contents of a folder may
be further structured in any way the application programmer sees
fit. Lately, TACOMA has included support for one specific type of
folder structure, namely the list structure, see
.IR folder_rotate (3C) 
and 
.IR folder_shift (3C). 
Unlike briefcases,
folders need not be created before they are used. They are
created auto-magically when data is put into them for the first time,
see 
.IR folder_store (3C).

.SH AGENTS
Agents are divided into two types of agents in this
implementation of TACOMA. The first type is the \fBserver agent\fR. It
provides services to the second type of agent, and are usually
bound to a site. The second type is the \fBclient agent\fR. It has the
ability to move between sites upon its own decision, and is able to
call upon the services of server agents (this is called a
meeting between agents in TACOMA).

.SS Server agents
Server agents are agents which can be the target of a
meeting performed by another agent (client). Upon a meeting
between a client agent and a server agent the client
agent passes it's briefcase to the server agent.

Server agents written in Tcl/Tk, other extensions to Tcl, or any
other programming language must be created as a script or binary
executable and placed in the \fC$TACOMA/sysagents\fR directory.
Briefcases are made available to the server agent in archived form
(see 
.IR archive (3C)) 
on the standard input. The archived briefcase is terminated by EOBC, which is
the 3 byte sequence $0A040A (linefeed, end of transmission, and a linefeed). The modified
briefcase should be archived and written out to standard output
(again terminated by EOBC), when the server agent is
about to return.

Server agents written in C can be linked with \fCc_stub.c\fR. This file
contains a main routine which handles briefcase passing from stdin/stdout.
\fCC_stub.c\fR is located
in the \fC$TACOMAPATH/cabinets/C_files\fR directory. Simply replace the string
\'\fB%CMAIN%\fR' in \fCc_stub.c\fR with the procedure to be called when the briefcase is
ready.

Agents written in the Tcl (or Tcl/Tk) language, can be written as Tcl
library routines and placed in the \fC$TACOMAPATH/cabinets/Tcl_files\fR
directory. In order to make such agents visible to the TACOMA system,
create a softlink from the sysagents directory pointing to
\fCtcl_stub.tcl\fR (or \fCtk_stub.tcl\fR) in the
\fC$TACOMAPATH/cabinets/Tcl_files\fR directory.
For instance, the ag_tcl agent is implemented as a Tcl library procedure
located in the \fC$TACOMAPATH/cabinets/Tcl_files\fR directory. It was made
visible to the TACOMA system by creating the following softlink:
.CS
ln -s $TACOMAPATH/cabinets/Tcl_files/tcl_stub.c $TACOMAPATH/sysagents/ag_tcl
.CE
Please remember that after adding a new agent in this way, you will need to
update the tclIndex file. This can be done by issuing the following command
when located in the Tcl_files directory:
.CS
echo "auto_mkindex . *.tcl" | tclsh
.CE
All agents written using one of the stub routines mentioned above are
written as procedures taking one argument, namely the briefcase passed to it. An example is
the agent  which executes Tcl code:
.CS
proc ag_tcl {bc} {
  upvar $bc passed_briefcase
  #
  # Operations may now be performed
  # on the briefcase 'passed_briefcase'
  #
}
.CE
.TP 8
\fBNotice\fR:
The argument \fIbc\fR and the local name \fIpassed_briefcase\fR
are completely arbitrary.
.SH ""
A server agent should also return a status code of zero
indicating success or a negative number indicating failure. When the
server returns a failure it may optionally supply further information
in the STATUS folder of the returned briefcase.

.SS Client agents
A client agent is an agent which calls upon a server
agent, delivering a briefcase in the process. This is
called a meeting between the two agents, and is accomplished
using the meet primitive, see
.IR meet (3C).
I.e. for an agent to meet with 'ag_tcl' it may use a call
in Tcl like this:
.CS
meet ag_tcl my_bc
.CE
Upon startup a client agent must first create a briefcase,
unless it is itself an server agent. The following
example shows the startup code for a client agent:
.CS
#!/usr/local/bin/tclsh
set auto_path [linsert $auto_path 0 $env(TACOMAPATH)/cabinets/Tcl_files]

bc_create my_bc
#
# Operations may now be performed
# on the briefcase 'my_bc'
#
.CE
.TP 8
\fBNotice\fR:
The first line of this example may need to be changed to reflect the
correct paths on your system.
.SH ""
We now proceed by giving some examples on how to create and activate
agents. These examples are all written in Tcl, but could just as well be
translated into any other language supported by TACOMA (although with a slightly
different syntax).

.SH EXAMPLE 1
We intend to write an agent which is started from any UNIX
shell and takes two arguments, a site and data to be output at the
console window of that site.
.CS
#!/usr/local/bin/tclsh
set auto_path [linsert $auto_path 0 $env(TACOMAPATH)/cabinets/Tcl_files]

bc_create my_bc
folder_store my_bc HOST [lindex $argv 0] ""
folder_store my_bc RETURNHOST $env(HOST)
folder_store my_bc DATA [lindex $argv 1]
folder_store my_bc OUTPUT "/dev/console"
meet ag_echo my_bc
exit
.CE
Using bc_create, we first create a new briefcase named my_bc.
Then, we specify on which host the meeting should take place by placing
a host name into the HOST folder. This host name is fetched from
the first argument of the script. Next, we specify a host which the
agent should report to when a meeting fails. We find the local machine
(point of origin) as the most appropriate place, and store it into the
RETURNHOST folder. The next two folders, DATA and OUTPUT,
are arguments to the server agent we wish to meet. They are described
in 
.IR ag_echo (1). 
Up to this point, all we have done is to prepare the
briefcase. Now, we simply need to initiate the meeting and
specify which agent we wish to meet with. Now if TACOMA is running on
the destination site, the agent should travel there and meet with
ag_echo.

.SH EXAMPLE 2

We intend to write an agent which is started from any UNIX
shell. It takes two arguments; a site and the filename for Tcl code to
be executed at that site.

file: My_agent
.CS
#!/usr/local/bin/tclsh
set auto_path [linsert $auto_path 0 $env(TACOMAPATH)/cabinets/Tcl_files]

bc_create my_bc
folder_store my_bc HOST [lindex $argv 0] ""
folder_store my_bc RETURNHOST $env(HOST)
set file [open [lindex $argv 1] r]
folder_store my_bc MYCODE [read -nonewline $file]
close $file
folder_store my_bc TCLCODE "MYCODE" ""
folder_store my_bc TCLMAIN "main" ""
meet ag_tcl my_bc
puts [folder_fetch my_bc STATUS]
exit
.CE
Example of Tcl code which could be stored into the MYCODE folder in the
previous example, also see 
.IR ag_tcl (1).

file: My_RemoteCode
.CS
proc main {bc} {
  upvar $bc loc_bc
  set error [catch {open "/etc/motd" r} file]
    if { $error == 0 } {
        folder_store loc_bc DATA [read -nonewline $file]
        close $file
    } else {
        folder_store loc_bc DATA "No message of the day at host"
    }
    folder_store loc_bc OUTPUT "/dev/console"
    meet ag_echo loc_bc
    return 0
}
.CE
If the file 'My_agent' (from above) has the executable flags set the
invocation:
.CS
My_agent "dslab3" My_RemoteCode
.CE
Will cause this effect
.TP
1)
The file 'My_RemoteCode' is stored into the MYCODE folder.
.TP
2)
The agent travels to the site named in the HOST folder.
This is the first argument 'dslab3'.
.TP
3)
There it meets with 'ag_tcl'
.TP
4)
Ag_tcl will take the agent code stored in the MYCODE
folder and execute the routine named in the TCLMAIN folder
(main).
.TP
5)
The agent puts the file '/etc/motd' into the DATA folder.
.TP
6)
The agent then travels back to the point of origin (named in the
RETURNHOST folder), because the HOST folder is exhausted (see
.IR meet (3C)).
.TP
7)
Here it meets with the agent called ag_echo.
.TP
8)
ag_echo puts the contents of the DATA folder into the console
window (see 
.IR ag_echo (1)).

.SH SEE ALSO
.IR tac_firewall (1), meet (3C), folder_store (3C), 
.IR bc_create (3C), cabinet_create (3C), archive (3C), ag_tcl (1), ag_c (1), ag_echo (1)

.SH AUTHOR
Nils Peter Sudmann (nilss@stud.cs.uit.no), Kjell Irgens (kjelli@stud.cs.uit.no) and Magne Rasmussen (magner@stud.cs.uit.no), University of Tromsoe.
Concept by Dag Johansen (dag@cs.uit.no), Robbert van Renesse
(rvr@cs.cornell.edu) and Fred B. Schneider (fbs@cs.cornell.edu).
