.TH folder_fetch 3C "TACOMA 1.1" "C Primitives"
.SH NAME
folder_fetch \- retrieves data from a folder 
.sp
folder_look \- reference data from a folder 
.sp
folder_storeA \- stores data into a folder 
.sp
folder_store \- varargs stub for folder_storeA 
.sp
folder_delete \- deletes a folder 
.SH SYNOPSIS
\fC#include <tacoma.h>\fR
.PP
\fCchar **folder_fetch(briefcase bc, char *name);\fR
.PP
\fCchar **folder_look(briefcase bc, char *name);\fR
.PP
\fCint folder_storeA(briefcase, char *name, char **mdata);\fR
.PP
\fCint folder_store(briefcase, char *name, char *data, ... );\fR
.PP
\fCvoid folder_delete(briefcase bc, char *name);\fR
.SH DESCRIPTION
These functions (except \fCfolder_look()\fR) are overloaded to work on both briefcases and cabinets. Specifically the old \fCcabinet_fetch()\fR, \fCcabinet_store()\fR and \fCcabinet_delete()\fR are no longer available in the C-API. 

\fCfolder_fetch()\fR retrieves information from the folder
.I name
in the briefcase (or cabinet)
.IR bc .
It will return a NULL-terminated array of strings that represents the
contents of the folder. It is the responsibility of the caller to free this array and the strings in it. For this, \fCtacoma.h\fR defines a macro called \fCfree_copy(char **)\fR which does exactly this.  

\fCfolder_look()\fR returns a reference pointing directly into the
internal structures of a briefcase. This is a lot more efficient than
\fCfolder_fetch()\fR, since no copy is created which later needs to be
discarded. However, notice that this reference could become invalid
after any one of \fCfolder_store()\fR, \fCfolder_delete()\fR,
\fCbc_discard()\fR, \fCunarchive()\fR, or \fCmeet()\fR has modified the
briefcase.   

\fCfolder_store()\fR
stores some information into a folder
of the briefcase (or cabinet) 
.IR bc .
It will either overwrite old information
if the folder already exists or create a new folder with
the supplied data. When using the varargs version 
\fCfolder_store()\fR the argument list 
.I data, ...
\fBmust\fR be terminated with
a NULL pointer. Contents can also be given to \fCfolder_storeA()\fR 
as an NULL terminated array of strings
.IR mdata .
If there is no data to store, \fCfolder_store()\fR will do nothing and
return -1.

\fCfolder_delete()\fR will delete a folder
.I name
and its content from the briefcase
.IR bc .
To delete an entire briefcase, see 
.IR bc_discard (3C).
.SH RETURN_VALUE
Upon successful operation a array of strings will be returned from
\fCfolder_fetch()\fR and \fCfolder_look()\fR. In case of failure or if the folder named 
.I name
does not exist, NULL will be returned. 
\fCfolder_store()\fR will return 0 on success and -1 on failure.
.SH BUGS 
\fCfolder_store()\fR and \fCfolder_fetch()\fR do not perform sanity checks on all memory allocations.  
.SH SEE ALSO
.IR bc_create (3C), folder_shift (3C), archive (3C).
.SH AUTHOR
Kjell Irgens (kjelli@stud.cs.uit.no), Magne Rasmussen (magner@stud.cs.uit.no)  and Nils Peter Sudmann (nilss@stud.cs.uit.no), University of
Tromsoe.
Concept by Dag Johansen (dag@cs.uit.no), Robbert van Renesse
(rvr@cs.cornell.edu) and Fred B. Schneider (fbs@cs.cornell.edu).
