'\"     # CS - begin code section
.de CS
.RS 4
\fC
.ps -1
.nf
.sp
..
'\"     # CE - end code section
.de CE
.fi
\fR
.ps +1
.RE
.sp .5
..
.TH folder_shift 3C "TACOMA 1.1" "C primitives"
.SH NAME
folder_shiftA \- shift the elements of a folder list
.sp
folder_shift \- varargs stub for folder_shiftA
.sp
folder_rotate \- rotate the elements of a folder list
.sp
folder_pop \- retrieves the first element from a folder list
.sp
folder_add \- add a element to the end of a folder list
.sp
folder_push \- add a element to the start of a folder list
.sp
folder_top \- reads the element a folder list
.SH SYNOPSIS
\fC#include <tacoma.h>\fR
.PP
\fCchar *folder_shiftA(briefcase bc, char *folder, enum direction, int amount, char **mdata );\fR
.PP
\fCchar *folder_shift(briefcase bc, char *folder, enum direction, int amount, char *data, ...);\fR
.PP
\fCchar *folder_rotate(briefcase bc, char *folder, enum direction, int amount);\fR
.PP
\fCchar *folder_pop(briefcase bc, char *folder)\fR
.PP
\fCchar *folder_add(briefcase bc, char *folder, char *data);\fR
.PP
\fCchar *folder_push(briefcase bc, char *folder, char *data);\fR
.PP
\fCchar *folder_top(briefcase bc, char *folder);\fR
.SH DESCRIPTION
In the TACOMA system the first element (or top if you think
about them as stacks) is the leftmost element of the
list. Consequently, the last element (or bottom in the case of stacks)
is the rightmost element. This affects all lists used by the system. Private
\fIfolder\fRs may follow any application programmer convention.

\fCfolder_shift()\fR shifts the elements of a \fIfolder\fR in one of two
\fIdirection\fRs, either \fCTAC_LEFT\fR or \fCTAC_RIGHT\fR. A shift removes a number of
elements equal to the \fIamount\fR given from the list. In the case of
a left shift, elements will be removed from the beginning of the
list. In the case of a right shift, elements will be removed from the
end of the list. The last element to be removed is the result of this
function, all other elements removed are lost. It is the responsibility of the caller to deallocate the string returned.

When using the varargs version 
\fCfolder_shift()\fR the argument list 
.I data, ...
\fBmust\fR be terminated with
a NULL pointer.
Optional \fIdata\fR can also be given to \fCfolder_shiftA()\fR 
as an NULL terminated array of strings
.IR mdata .
If the optional \fIdata\fR
elements are supplied, they will be added to the list from the
opposite \fIdirection\fR. The \fIdata\fR will be added to the
\fIfolder\fR in the same order as they appear on the command line. So,
if the \fIdirection\fR of the shift is right, the rightmost element of
the \fIdata\fR is added first. If the \fIdirection\fR is left,
the leftmost \fIdata\fR is added first.

If the \fIamount\fR to be
shifted is a negative number, the direction changes from left to right
or right to left. If the \fIamount\fR to be shifted is zero, no shift is
performed. The function will return the current leftmost or rightmost
element. If the folder does not exist, \fCfolder_shift()\fR will create a new
folder if any \fIdata\fR is supplied. It will do nothing if no \fIdata\fR
is supplied. If the \fIfolder\fR becomes empty as a result of the shift it 
is silently removed. In any case, it will return an NULL pointer if the folder
does not exist.

\fCfolder_rotate()\fR rotates the elements of a \fIfolder\fR in one of
two \fIdirection\fRs, either \fCTAC_LEFT\fR or \fCTAC_RIGHT\fR. A number of elements equal
to the \fIamount\fR given will be rotated. A rotation will move
elements from the left or the right end of a list to the other
end. The last element to be rotated is the result of this function. It is the responsibility of the caller to deallocate this result.
No elements are lost in the rotation. If the \fIamount\fR to be rotated
is a negative number, the direction changes from left to right or
right to left. If the \fIamount\fR to rotated is zero or the folder does not
exist, no rotation is
performed and the function returns a NULL pointer.
.SS Stub routines
The following functions are simple defines using \fCfolder_shift()\fR.
They are kept because their syntax is more intuitive than
\fCfolder_shift()\fR. If you feel confident in your understanding of the
\fCfolder_shift()\fR, you should use the mappings directly. They are
as follows:

.RS
folder_pop ... = folder_shift ... left 1
.br
folder_add ... data = folder_shift ... left 0 data
.br
folder_push ... data = folder_shift ... right 0 data
.br
folder_top ... = folder_shift ... left 0
.RE

\fCfolder_pop()\fR retrieves the first element from a folder
list. It will also remove this element from the \fIfolder\fR list. If
the \fIfolder\fR becomes empty, it is deleted.

\fCfolder_add()\fR adds an element to the end of a folder list. It will
either add the given data as a single element to already existing
elements of the \fIfolder\fR, or create a new \fIfolder\fR
with the supplied data as a single element if the \fIfolder\fR does not
exist.
By using this function and \fCfolder_pop()\fR, one can implement queues
(FIFO) using a single folder.

\fCfolder_push()\fR adds an element to the start of a folder list. It will
either add the given data as a single element to already existing
elements of the \fIfolder\fR, or create a new \fIfolder\fR with the supplied data
as a single element if the \fIfolder\fR does not exist.
By using this function and \fCfolder_pop()\fR, can implement stacks
(LIFO) using a single folder.

\fCfolder_top()\fR reads the first element from a folder list. Unlike
\fCfolder_pop()\fR, it will \fBnot\fR remove this item from the \fIfolder\fR.

.SH BUGS
These functions do \fBnot\fR work on cabinets, yet.

.SH SEE ALSO
.IR folder_store (3C), bc_create (3C)

.SH AUTHOR
Kjell Irgens (kjelli@stud.cs.uit.no), Magne Rasmussen
(magner@stud.cs.uit.no) and Nils Peter Sudmann (nilss@stud.cs.uit.no), University of
Tromsoe.
Concept by Dag Johansen (dag@cs.uit.no), Robbert van Renesse
(rvr@cs.cornell.edu) and Fred B. Schneider (fbs@cs.cornell.edu).
