/*******************************************************************************
 *
 * $Source: /users/nilss/source/RCS/tac_firewall.c,v $
 * $Header: tac_firewall.c,v 1.8 96/04/19 18:35:18 nilss Exp $
 * $Author: nilss $
 * $Date: 96/04/19 18:35:18 $
 * Language: C
 * Status: Experimental
 *
 * (c) Copyright 1996, TACOMA, all rights reserved.
 *
 ******************************************************************************/
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "tacoma.h"
#include "meet.h"
#include "secret.h"
#include "version.h"

/*
 * Tacoma version.
 */
#define	TAC_VERSION	"TACOMA v"TACOMA_VERSION



/*
 * Use DEBUG to debug firewall.
 */
/*#define DEBUG*/

#ifndef SA_RESTART
#  define	SA_RESTART	0
#endif /* SA_RESTART */
#ifndef HAVE_SETENV
#  define setenv(x,y,z)	(putenv(strdup((TAC_ENV_PATH "=/"))))
#endif /* HAVE_SETENV */
#ifndef INADDR_NONE
#  define INADDR_NONE (-1)
#endif /* INADDR_NONE */
#ifndef MAXHOSTNAMELEN
#  define MAXHOSTNAMELEN 256
#endif /* MAXHOSTNAMELEN */

/*
 * Option flags.
 */
#define	FWO_CHROOT	(1<<0)
#define	FWO_NODAEMON	(1<<1)
#define	FWO_VERBOSE	(1<<2)

/*
 * Messages written by firewall.
 */
#define	FWM_WELCOME	("0 Welcome" TM_BCEND)
#define	FWM_ACCEPT 	("0 Got it" TM_BCEND)
#define	FWM_NOACCEPT	("2 Only signed briefcases are accepted " \
			"from your host" TM_BCEND)
#define	FWM_NOGOODSIGN	("2 Your signature is no good at this host" TM_BCEND)
#define	FWM_ABORTERR	("6 Internal failure, abort" TM_BCEND)


/*
 * Global variables used by firewall.
 */
static string		program;
static string*		acceptfrom;
static unsigned long*	autoaccept;


/*
 * Utility functions used in this file.
 */
static void		tac_firewall( int, int );
static int		tac_get_env( string* );
static int		tac_parse( int, string* );
static unsigned long*	tac_autoaccept( cabinet, int );
static void		tac_autostart( cabinet, string, int );
static void		tac_atexit( void );
static void		tac_sigterm( int );
static void		tac_sigchld( int );
static string		tac_newroot( string, int options );


int
main( int argc, char* argv[] )
{
    int			port;
    int			sockfd;
    int			options;
    char		tacomahost[MAXHOSTNAMELEN];
    time_t		start_time;
    string		tacomapath;
    cabinet		inlog;
    cabinet		outlog;
    cabinet		syscab;
    struct sockaddr_in	addr;

    /* remember program name */
    for ( program=argv[0]; strchr(program, '/'); program++ )
	program = strchr(program, '/');
    program = strdup(program);

    /* get our hostname */
    gethostname(tacomahost, MAXHOSTNAMELEN);

    /* get values from environment */
    port = tac_get_env(&tacomapath);

    /* parse command line options */
    options = tac_parse(argc, argv);

    /* print banner */
    printf("\n" TAC_VERSION "\n");
    printf("\nFirewall started at %s (on port %d)\n", tacomahost, port);
    start_time = time(NULL);
    printf("Date: %s", ctime(&start_time));
    printf("Path: %s\n", tacomapath);

    /* clear file mode creation mask */
    umask(0);

    /* change working directory */
    if ( chdir(tacomapath) ) {
	fprintf(stderr, "%s: Unable to access %s: %s\n",
		program, tacomapath, strerror(errno));
	exit(1);
    }

    /* change root if user has requested this */
    if ( options & FWO_CHROOT )
	tacomapath = tac_newroot(tacomapath, options);

    /* check system cabinets, create if nonexisting */
    inlog = cabinet_open(TAC_INLOG_CAB, TRUE);
    outlog = cabinet_open(TAC_OUTLOG_CAB, TRUE);
    syscab = cabinet_open(TAC_SYSTEM_CAB, TRUE);

    if ( inlog == NULL || outlog == NULL || syscab == NULL ) {
	fprintf(stderr, "\n%s: Could not access system cabinets: %s\n",
		program, strerror(errno));
	exit(1);
    }

    /* have not use for log cabinets here */
    cabinet_close(inlog);
    cabinet_close(outlog);

    /* find which persons we accept agents from */
    acceptfrom = folder_fetch(syscab, TAC_ACCEPTFROM);

    /* add trusted hosts */
    autoaccept = tac_autoaccept(syscab, options);

    /* start up listening socket */
    sockfd = socket (AF_INET, SOCK_STREAM, 0);

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(port);

    if ( bind(sockfd, (struct sockaddr *) &addr, sizeof(addr)) == -1 ) {
	fprintf(stderr, "%s: bind failed: %s\n", program, strerror(errno));
	exit(1);
    }

    if ( listen(sockfd, 5) == -1 ) {
	fprintf(stderr, "%s: failed to listen to socket: %s\n",
		program, strerror(errno));
	exit(1);
    }

    /* run as daemon if not user has told otherwise */
    if ( !(options & FWO_NODAEMON) ) {
	switch ( fork() ) {
	  case -1:
	    fprintf(stderr, "%s: could not fork: %s\n",
		    program, strerror(errno));
	    exit(1);

	  case 0:
	    /* child continues */
	    if ( options & FWO_VERBOSE )
		printf("Daemon pid is %d\n", (int) getpid());
	    break;

	  default:
	    /* parent goes bye-bye, but sleep just a while first */
	    sleep(2);
	    exit(0);
	}

	/* become session leader */
	if ( setsid() == -1 ) {
	    fprintf(stderr, "%s: could not become session leader: %s\n",
		    program, strerror(errno));
	    exit(1);
	}
    }

    /* autostart agents */
    tac_autostart(syscab, tacomapath, options);

    /* no more use for system cabinet here */
    cabinet_close(syscab);

    /* call main firewall routine */
    tac_firewall(sockfd, options);

    /* should never get here */
    return (1);
}


static void
tac_firewall( int sockfd, int options )
{
    int			fd;
    int			idx;
    int			a_len;
    int			flags;
    int			retval;
    char		status[TM_STAT_MSG_SIZE];
    string		arc;
    string		buffer;
    string		contact;
    string*		strings;
    cabinet		inlog;
    briefcase		bc;
    struct sockaddr_in	a_from;

    /* open logging cabinet */
    if ( (inlog = cabinet_open(TAC_INLOG_CAB, TRUE)) == NULL ) {
	fprintf(stderr, "\n%s: Could not access inlog cabinet: %s\n",
		program, strerror(errno));
	exit(1);
    }

    /* enter connect accept loop */
    while ( TRUE ) {
	/* wait for contact with remote meet */
	a_len = sizeof(struct sockaddr_in);
	while ((fd = accept(sockfd, (struct sockaddr *)&a_from, &a_len)) == -1)
	    if ( errno == EINTR )
		continue;
	    else {
		fprintf(stderr, "%s: failed to accept client connection: %s\n",
			program, strerror(errno));
		exit(1);
	    }

#ifdef DEBUG
	fprintf(stderr, "Accepting a bc from %s\n",
		inet_ntoa(a_from.sin_addr));
#endif

	/* fork off child to handle meet */
	switch ( fork() ) {
	  case -1:
	    fprintf(stderr, "%s: could not fork: %s\n",
		    program, strerror(errno));
	    exit(1);

	case 0:
	    /* close listening socket */
	    close(sockfd);

	    /* write welcome message */
            write(fd, FWM_WELCOME, sizeof(FWM_WELCOME)-1);
#ifdef DEBUG
	    fprintf(stderr, "Wrote acceptance, ready to recieve the bc\n");
#endif

	    /* read flags and briefcase */
	    buffer = tac_read_bc_stream(fd);

	    /* check flags */
	    if ( isdigit(*buffer) )
		flags = atoi(buffer);
	    else {
		write(fd, FWM_ABORTERR, sizeof(FWM_ABORTERR)-1);
		free(buffer);
		close(fd);
		_exit(6);
	    }

	    /* should we accept briefcase? */
	    if ( !(flags & TM_SIGN) ) {
		for ( idx=0; autoaccept[idx] != INADDR_NONE; idx++ )
		    if ( autoaccept[idx] == a_from.sin_addr.s_addr )
			break;
		if ( autoaccept[idx] == INADDR_NONE ) {
		    write(fd, FWM_NOACCEPT, sizeof(FWM_NOACCEPT)-1);
		    free(buffer);
		    close(fd);
		    _exit(2);
		}
	    }

	    /* unpack received briefcase */
	    bc = bc_create();
	    arc = strchr(buffer, ' ') + 1;
	    unarchive(bc, arc);

	    /* log incoming agent */
	    if ( !(flags & TM_SYNC) &&
		 ( !(flags & TM_NLOG) || (flags & (TM_CRPT|TM_SIGN))) )
		folder_store(inlog, *folder_look(bc, TAC_AGENTID), arc, NULL);
	    free(buffer);

	    /* set uid to REAL uid (not effective) */
	    if ( setuid(getuid()) == -1 ) {
		write(fd, FWM_ABORTERR, sizeof(FWM_ABORTERR)-1);
		close(fd);
		_exit(6);
	    }

	    /* see if we have to decrypt and/or check signature */
	    if ( flags & (TM_CRPT|TM_SIGN) ) {
		/* fill folder with info for crypt agent */
		folder_store(bc, TAC_CR_PASSPHRASE, PASSPHRASE, NULL);
		folder_store(bc, TAC_CR_DECRYPT, TAC_CR_BC, NULL);
		if ( (strings = folder_look(bc, TAC_CR_METHOD)) )
		    folder_add(bc, TAC_CR_DECRYPT, *strings);

		/* meet with crypt agent */
		if ( meet(TAC_AG_CRYPT, bc, TM_LOCL|TM_SYNC) ) {
		    write(fd, "6 ", 2);
		    if ( (strings = folder_look(bc, TAC_STATUS)) )
			write(fd, *strings, strlen(*strings));
		    else
			write(fd, "Unknown error", 13);
		    write(fd, TM_BCEND, TM_BCEND_SIZE);
		    close(fd);
		    _exit(6);
		}

		/* is signer one we trust? */
		strings=folder_look(bc, TAC_CR_BCSIGN);
		for ( idx=0;
		      acceptfrom[idx] && strcmp(acceptfrom[idx], *strings);
		      idx++ )
		    ;

		if ( acceptfrom[idx] == NULL ) {
		    /* bad siger */
		    write(fd, FWM_NOGOODSIGN, sizeof(FWM_NOGOODSIGN)-1);
		    close(fd);
		    _exit(2);
		}

		strings = folder_fetch(bc, TAC_CR_BC);
		bc_discard(bc);
		bc = bc_create();
		unarchive(bc, *strings);
		free(strings);
	    }

	    /* briefcase OK, write acceptance */
	    write(fd, FWM_ACCEPT, sizeof(FWM_ACCEPT)-1);

	    /* end of communication unless sync meet */
	    if ( !(flags & TM_SYNC) )
		close(fd);

	    /* meet with contact */
	    contact = folder_rotate(bc, TAC_CONTACT, TAC_LEFT, 1);
#ifdef DEBUG
	    fprintf(stderr, "Wrote log, replied OK. Now performing a %s " \
		    "local meet with %s.\n",
		    (flags & TM_SYNC) ? "sync" : "async", contact);
#endif
	    retval = meet(contact, bc, flags|TM_LOCL);
	    free(contact);

	    if ( flags & TM_SYNC ) {
#ifdef DEBUG
		fprintf(stderr, "Returned bc\n===========\n%s\n===========\n",
			arc = archive(bc));
		free(arc);
#endif
		sprintf(status, "%d ", retval);
		write(fd, status, strlen(status));
		tac_write_bc_stream(fd, bc);
	    }

	    /* finished meet, cleanup */
	    close(fd);
	    _exit(0);

	default:
	    /* close connection socket */
	    close(fd);
	    break;
	}
    }
}


/*
 * Utility functions used in this file.
 */
static int
tac_get_env( string* path )
{
    int	    port;
    string  tacomapath;
    string  tacomaport;

    /* get and check environment variables */
    tacomapath = getenv(TAC_ENV_PATH);
    tacomaport = getenv(TAC_ENV_PORT);
    if ( tacomaport && isdigit(*tacomaport) )
	port = atoi(tacomaport);
    else
	port = -1;

    /* write error message and exit if any errors */
    if ( tacomapath == NULL || tacomaport == NULL || port == -1 ) {
	if ( tacomapath == NULL )
	    fprintf(stderr, "%s: Environment variable '" TAC_ENV_PATH \
		    "' not set.\n", program);
	if ( tacomaport == NULL )
	    fprintf(stderr, "%s: Environment variable '" TAC_ENV_PORT \
		    "' not set.\n", program);
	else
	    if ( port == -1 )
		fprintf(stderr, "%s: Environment variable '" TAC_ENV_PORT \
			"' must be a number.\n", program);
	exit(1);
    }

    /* return environment values */
    *path = tacomapath;

    return (port);
}

static int
tac_parse( int argc, string* argv )
{
    int	 idx;
    int	 usage;
    int	 options;

    usage = FALSE;
    options = 0;

    for ( idx=1; idx<argc; idx++ )
	if ( (argv[idx][0] == '-') && (argv[idx][2] == '\0') )
	    switch ( argv[idx][1] ) {
	      case 'd':
		options |= FWO_NODAEMON;
		break;
	      case 'r':
		options |= FWO_CHROOT;
		break;
	      case 'v':
		options |= FWO_VERBOSE;
		break;
	      default:
		usage = TRUE;
	    }
	else
	    usage = TRUE;

    if ( usage ) {
	fprintf(stderr, "\nUsage: %s [-r] [-d] [-v]\n", program);
	fprintf(stderr, "\t-r Change filesystem root\n");
	fprintf(stderr, "\t-d Don't run firewall as daemon\n");
	fprintf(stderr, "\t-v Be verbose\n");
	fprintf(stderr, "\n");
	exit(1);
    }

    return (options);
}

static unsigned long*
tac_autoaccept( cabinet syscab, int options )
{
    int		     idx;
    int		     count;
    string*	     str_p;
    string*	     strings;
    unsigned long*   hosts;
    struct hostent*  hep;

    printf("Adding trusted hosts%c", (options & FWO_VERBOSE) ? '\n' : ' ');
    count = 0;
    hosts = NULL;

    /* all trusted hosts are listed in folder TAC_AUTOHOSTS */
    strings = folder_fetch(syscab, TAC_AUTOHOSTS);

    /* find inet-address of all trusted hosts */
    for ( str_p=strings; str_p && *str_p; str_p++ ) {
	if ( **str_p == '\0' )
	    continue;

	/* try hostname lookup */
	if ( (hep = gethostbyname(*str_p)) == NULL )
	    continue;

	/* found address of host, now add to list */
	for ( idx=0; hep->h_addr_list[idx]; idx++, count++ ) {
	    if ( !hosts ) 
		hosts = malloc((count+2)*sizeof(unsigned long));
	    else
		hosts = realloc(hosts, (count+2)*sizeof(unsigned long));
	    if ( hosts ) {
		hosts[count] =
		    ((struct in_addr *)(hep->h_addr_list[idx]))->s_addr;
		if ( options & FWO_VERBOSE )
		    printf("\t%s (%s)\n", hep->h_name,
			   inet_ntoa(*(struct in_addr *)
				     hep->h_addr_list[idx]));
		else
		    printf(".");
	    }
	    else {
		fprintf(stderr, "%s: Memory deficiency.\n", program);
		exit(1);
	    }
	}
    }
    if ( hosts )
	hosts[count] = INADDR_NONE;

    /* now free data from folder */
    free_copy(strings);

    printf("%sdone (%d hosts added)\n",
	   (options & FWO_VERBOSE) ? "" : " ", count);

    return (hosts);
}

static void
tac_autostart( cabinet syscab, string path, int options )
{
    char	      cmd[MAXPATHLEN];
    string*	      str_p;
    string*	      strings;
    struct sigaction  act;

    printf("Autostarting agents%c", (options & FWO_VERBOSE) ? '\n' : ' ');

#ifdef HAVE_ATEXIT
    /* register function to kill all children if exiting */
    if ( atexit(tac_atexit) == -1 ) {
	fprintf(stderr, "%s: could not register exit handler: %s\n",
		program, strerror(errno));
	exit(1);
    }
#endif /* HAVE_ATEXIT */
    /* register signal handler to kill all children if killed */
    act.sa_handler = tac_sigterm;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if ( sigaction(SIGTERM, &act, NULL) == -1 ) {
	fprintf(stderr, "%s: could not register signal handler: %s\n",
		program, strerror(errno));
	exit(1);
    }

    /* register signal handler to catch dead children */
    act.sa_handler = tac_sigchld;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
    if ( sigaction(SIGCHLD, &act, NULL) == -1 ) {
	fprintf(stderr, "%s: could not catch dead children: %s\n",
		program, strerror(errno));
	exit(1);
    }

    /* start all agents listed in TAC_STARTUP_AG */
    strings = folder_fetch(syscab, TAC_STARTUP_AG);

    for ( str_p=strings; str_p && *str_p; str_p++ ) {
	if ( **str_p == '\0' )
	    continue;
	sprintf(cmd, "%s%s" TAC_BIN "%s", path,
		(path[strlen(path)-1] == '/') ? "" : "/", *str_p);

	/* start agent */
	if ( vfork() == 0 ) {
	    execl(cmd, *str_p, NULL);
	    fprintf(stderr, "%s: failed to start %s: %s\n", program,
		    cmd, strerror(errno));
	    _exit(1);
	}

	if ( options & FWO_VERBOSE )
	    printf("\t%s\n", *str_p);
	else
	    printf(".");
    }

    /* now free data from folder */
    free_copy(strings);

    printf("%sdone\n", (options & FWO_VERBOSE) ? "" : " ");
}

static void
tac_atexit( void )
{
    /* kill all processes in process group (all children) */
    kill(0, SIGTERM);
}

static void
tac_sigterm( int signo )
{
    /* tac_atexit will kill all children */
    exit(0);
}

static void
tac_sigchld( int signo )
{
    /* non-blocking wait for any child, discard status */
    waitpid(-1, NULL, WNOHANG);
}

static string
tac_newroot( string tacomapath, int options )
{
    string  path;

    /* set new root */
    if ( chroot(tacomapath) == -1 ) {
	fprintf(stderr, "%s: could not set new root: %s\n", program,
		strerror(errno));
	exit(1);
    }

    /* make sure working directory is at new root */
    if ( chdir("/") == -1 ) {
	fprintf(stderr, "%s: could not 'cd /': %s\n", program,
		strerror(errno));
	exit(1);
    }

    /* update environment to new situation */
    if ( setenv(TAC_ENV_PATH, "/", TRUE) == -1 ) {
	fprintf(stderr, "%s: setenv() failed: %s\n", program, strerror(errno));
	exit(1);
    }

    /* check that environment has ben updated correctly */
    if ( (path = getenv(TAC_ENV_PATH)) == NULL || strcmp(path, "/") != 0 ) {
	fprintf(stderr, "%s: environment variable '" TAC_ENV_PATH \
		"' not correctly set.\n", program);
	exit(1);
    }

    if ( options & FWO_VERBOSE )
	printf("Root set to %s\n", tacomapath);
    else
	printf("(new root)\n");

    return (path);
}
