/*******************************************************************************
 *
 * $Source: /users/nilss/source/RCS/tacoma.c,v $
 * $Header: tacoma.c,v 1.12 96/04/19 18:46:31 nilss Exp $
 * $Author: nilss $
 * $Date: 96/04/19 18:46:31 $
 * Language: C
 * Status: Experimental
 *
 * (c) Copyright 1996, TACOMA, all rights reserved.
 *
 ******************************************************************************/
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#if HAVE_DIRENT_H
# include <dirent.h>
#else /* not HAVE_DIRENT */
# define dirent direct
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif /* HAVE_DIRENT */
#include <fcntl.h>
#include <errno.h>
#ifndef HAVE_STRERROR
extern char *sys_errlist[];
extern int sys_nerr;
char *strerror(int error){
    if (error>0&&error<sys_nerr)
	return sys_errlist[error];
    else
	return NULL;
}
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif

#include "tacoma.h"

/*
 *	Utility functions used in this file.
 */
static int	cabinet_store( cabinet, string, string* );
static string*	cabinet_fetch( cabinet, string );
static int	tac_hashit( string );
static folder	tac_find_folder( briefcase, string );
static string*	tac_str2array( string );
static int       tac_init_cab_path( void );

static char cabpath[MAXPATHLEN]="";

#define tac_free_folder( f ) {\
  free_copy(f->contents);\
  free(f->name);\
  free(f);\
}
	 

/*****************************************************************************
 *
 *	Functions for creating and manipulating briefcases.
 *
 ****************************************************************************/
/*
 * Creates a new local briefcase. The briefcase should be deallocated
 * by bc_discard.
 *
 *   Returns:	new briefcase, or NULL if error.
 */
briefcase
bc_create( void )
{
    briefcase	bc;
    char	ag_id[TAC_ID_SIZE];
    static int	bccount = 0;

    /* allocate briefcase and hash buckets, free allocated memory and
     * return NULL if error.
     */
    if ( (bc = malloc(sizeof(briefcase_rec))) )
    {
	bc->name = NULL;
	if ( (bc->f = calloc(TAC_BC_HASHSIZE, sizeof(folder))) )
	{
	    /* create an agent ID and store in folder TAC_AGENTID */
	    sprintf(ag_id, "%ldl%d%d", time(NULL), (int)getpid(), bccount++);
	    if ( folder_store(bc, TAC_AGENTID, ag_id, NULL) == 0 )
		return (bc);

	    free(bc->f);
	}
	free(bc);
    }
    return (NULL);
}

/*
 * Deletes a local briefcase. The briefcase 'bc' and all folders associated
 * with it are deallocated, and should not be referenced again.
 *
 *   Input:	bc	The briefcase to delete.
 */
void
bc_discard( briefcase bc )
{
    int	    idx;
    folder  f, next;

    /* sanity checks */
    if ( bc && bc->name ) {
	fprintf(stderr, "bc_discard(): Trying to discard cabinet");
	return;
    }

    if ( (bc == NULL) || (bc->f == NULL) )
	return;

    /* find and deallocate memory used by all folders */
    for ( idx=0; idx<TAC_BC_HASHSIZE; idx++ )
	for ( f=bc->f[idx]; f; f=next )
	{
	    next = f->next;
	    tac_free_folder(f);
	}

    /* deallocate name, hash buckets and briefcase */
    free(bc->name);
    free(bc->f);
    free(bc);
}

/*
 * Lists the folders of a local briefcase. The array of strings and the
 * strings themselves should be deallocated with 'free' after use. (Or
 * use the macro 'free_copy(sa)' to free it.)
 *
 *   Input:	bc	The briefcase to list folders from.
 *
 *   Returns:	A NULL-terminate array of strings that each represents the
 *		name of a folder in the briefcase, or NULL if error.
 */
string*
bc_list( briefcase bc )
{
    folder   f;
    string*  names;
    int	     i, count, folders=0;

    /* sanity checks */
    if ( bc && bc->name ) {
	fprintf(stderr, "bc_list(): Trying to list cabinet");
	return (NULL);
    }

    if ( (bc == NULL) || (bc->f == NULL) )
	return (NULL);

    /* find how many folders in briefcase */
    for ( i=0; i<TAC_BC_HASHSIZE; i++ )
	for ( f=bc->f[i]; f; f=f->next )
	    folders++;

    /* allocate array of strings */
    if ( (names = malloc(sizeof(string)*(folders+1))) == NULL )
	return (NULL);
    names[count=folders] = NULL;

    /* fill the array with names of folders */
    for ( i=0; i<TAC_BC_HASHSIZE; i++ )
	for ( f=bc->f[i]; f; f=f->next )
	    if ( (names[--count] = strdup(f->name)) == NULL )
	    {
		/* free any allocated memory if strdup fails */
		for ( ; count<folders; count++ )
		    free(names[count]);
		free(names);
		return (NULL);
	    }

    return (names);
}

/*
 * Archive a briefcase. Packs the passed briefcase and returns it in the
 * form of a string. The user is responsible for deallocating the
 * returned string.
 *
 *   Input:	bc	The briefcase to archive.
 *
 *   Returns:	The packed briefcase as a string, or NULL if error.
 *
 *   Feature?:	Will allocate one character more than necessary. This is
 *		done to simplify insertion of spaces.
 *
 *   Bugs:	What happens if a data element starts with, or contains
 *		a '{' ?
 */
string
archive( briefcase bc )
{
    int	     idx;
    int	     len;
    int      escps;
    int	     alloclen;
    folder   f;
    string   str;
    string   tstr;
    string   arcstr;
    string*  strings;

    /* sanity checks */
    if ( bc && bc->name ) {
	fprintf(stderr, "archive(): Trying to archive cabinet");
	return (NULL);
    }

    if ( (bc == NULL) || (bc->f == NULL) )
	return (NULL);

    /* find length of archive string */
    alloclen = 0;
    for ( idx=0; idx<TAC_BC_HASHSIZE; idx++ )
	for ( f=bc->f[idx]; f; f=f->next ) {
	    alloclen += strlen(f->name) + 1;
	    if ( *f->contents && (f->contents)[1] )
		alloclen += strlen(f->name) + 7;
	    for ( strings=f->contents; *strings; strings++ ) {
		len = strlen(*strings);
		for (escps=0, tstr=*strings; *tstr; tstr++) 
		    switch (*tstr) {
		    case '{': 
		    case '}': 
		    case '\\': 
			escps++; 
			break; 
		    default: 
			break;
		    }
		alloclen += escps;
		alloclen += (len == 0 ? 3 : len+1);
		if ( strchr(*strings, ' ') || **strings=='{' )
		    alloclen += 2;
	    }
	}

    /* allocate memory for archive string */
    arcstr = str = malloc(alloclen+1);

    /* process each folder */
    for ( idx=0; idx<TAC_BC_HASHSIZE; idx++ )
	for ( f=bc->f[idx]; f; f=f->next ) {
#ifdef SANITY
	    if ( !f->contents || !*f->contents ) {
		fprintf(stderr, "Empty folder passed to archive().\n");
		fprintf(stderr, TAC_NOTIFY);
		break;
	    }
#endif /* SANITY */
	    if ( *f->contents && (f->contents)[1] ) {
		sprintf(str, "*#%s 1 %s {", f->name, f->name);
		str += strlen(str);
	    } else {
		strcpy(str, f->name);
		str += strlen(str);
		*str++ = ' ';
	    }
	    for ( strings=f->contents; *strings; strings++ )
		if ( **strings ) {
		    if (strchr(*strings, ' ')) *str++ ='{';
		    for (tstr=*strings; *tstr;) {
			switch (*tstr) {
			case '{': 
			case '}': 
			case '\\':
			    *str++ = '\\';
			    break;
			default:
			    break;
			}
			*str++ = *tstr++;
		    }
		    if (strchr(*strings, ' ')) *str++ ='}';
		    *str++ = ' ';
		} else {
		    strcpy(str, "{} ");
		    str += 3;
		}
	    if ( *f->contents && (f->contents)[1] ) {
		strcpy(--str, "} ");
		str += 2;
	    }
	}
    *(--str) = '\0';

#ifdef SANITY
    if ( alloclen != strlen(arcstr)+1 )
    {
	fprintf(stderr, "archive(): alloclen = %d,  strlen(arcstr) = %d\n",
		alloclen, strlen(arcstr));
	fprintf(stderr, TAC_NOTIFY);
	if ( alloclen < strlen(arcstr)+1 ) {
	    fprintf(stderr, "\nExiting since allocated memory too small!\n\n");
	    exit (1);
	}
    }
#endif /* SANITY */

    return ( arcstr );
}

/*
 * Unarchive a briefcase. Unpacks the passed archive into the supplied
 * briefcase which is also returned. Since 'unarchive()' only adds/overwrites
 * folders found in the archive to the supplied briefcase this briefcase
 * doesn't have to be empty. However, folders existing in both the briefcase
 * and archive will be overwritten in the briefcase.
 *
 *   Input:	bc	The briefcase to unpack the archive into.
 *		archive	The archive to unpack.
 *
 *   Returns:	The passed briefcase, with the folders in the archive added,
 *		or NULL if error.
 *
 *   Bugs:	Might become ugly if data elements contain unbalanced braces.
 */
briefcase
unarchive( briefcase bc, string arc )
{
    int		size;
    int		leftbraces;
    string	str;
    string	data;
    string	s_ptr;
    string	name;
    string*	strings;
    string*	metas;
    string*	metalist;
    briefcase	meta;

    /* sanity checks */
    if ( bc && bc->name ) {
	fprintf(stderr, "unarchive(): Trying to unarchive cabinet");
	return (NULL);
    }

    if ( (bc == NULL) || (bc->f == NULL) || (arc == NULL) )
	return (NULL);

    /* used to find multipart folders */
    meta = bc_create();

    /* now start processing archive */
    for ( str=arc, name=NULL, leftbraces=0; *arc; str++ )
	switch ( isspace(*str) ? ' ' : *str ) 
	    {
	    case '{':
		leftbraces++;
		break;
	    case '}':
		leftbraces--;
		break;
	    case '\\':
		str++;
		break;
	    case '\n':
	    case ' ':
		if ( leftbraces )
		    break;
		if ( str == arc ) {
		    arc++;
		    break;
		}
	    case '\0':
		/* store element */
		if ( *arc == '{' )
		    size = str-(arc++)-1;
		else
		    size = str-arc+1;
		data = s_ptr = malloc((size_t) size--);
		for (; size>0; size--) {
		    if (*arc=='\\') {
			switch (arc[1]) {
			case '\\':
			case '{':
			case '}':
			    size--; arc++; break;
			default:
			    break;
			}
		    }
		    *s_ptr++ = *arc++; 
		}
	        *s_ptr='\0';
		if ( !name )
		    name = data;
		else {
		    if( *name=='*' && name[1]=='#' )
			folder_store(meta, &name[2], data, NULL);
		    else
			folder_store(bc, name, data, NULL);
		    free(name);
		    free(data);
		    name=NULL;
		}
		if ( *str == '\0' )
		    arc = str;
		else
		    arc = str+1;
	        break;
	    default:
		break;
	}

    /* Now check metafolders for multiparts... */
    for( metas=metalist=bc_list(meta); *metalist; metalist++ )
	if ( strcmp(TAC_AGENTID, *metalist) )
	{
	    strings=tac_str2array(*folder_look(bc, *metalist));
	    folder_storeA(bc, *metalist, strings);
	    free_copy(strings);
	}
    free_copy(metas);
    bc_discard(meta);

    return (bc);
}


/*****************************************************************************
 *
 *	Functions for creating and manipulating cabinets.
 *
 ****************************************************************************/
/*
 * Opens the (persistent) local cabinet 'name'. If the cabinet does not exist,
 * and the argument 'create' is TRUE, the cabinet will be created, otherwise
 * NULL is returned. The cabinet returned should be deallocated by
 * 'cabinet_discard()'.
 *
 *   Input:	name	Name of the cabinet.
 *		create	If TRUE, the cabinet will be created if it does not
 *			already exist.
 *
 *   Returns:	The open cabinet 'name', or NULL if error or the cabinet
 *		does not exist and 'create' is FALSE.
 */
cabinet
cabinet_open( string name, int flags )
{
    char	 path[MAXPATHLEN];
    cabinet	 cab;
    struct stat	 afs;

    /* sanity check */
    if ( !name || !*name )
	return (NULL);

    if (!*cabpath && !tac_init_cab_path()) return(NULL);
 
    strcpy(path, cabpath);
    strcat(path, name);

    /* check if cabinet exists */
    if ( stat(path, &afs) == -1 ) {
	/* cabinet does not exist, create? */
	if ( (errno == ENOENT) && (flags & TC_CREAT) ) {
	    if ( mkdir(path, 0755) == -1 ) {
		fprintf(stderr, "Could not create cabinet %s", path);
		return (NULL);
	    }
	}
	else
	    return (NULL);
    }
    else {
	/* cabinet exists, check permissions */
	if ( !(afs.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH)) )
	    return (NULL);
    }
    
    /* make cabinet */
    if ( (cab = (cabinet) malloc(sizeof(cabinet_rec))) ) {
	cab->f = NULL;
	if ( (cab->name = strdup(name)) )
	    return (cab);

	/* could not get memory, so free everything */
	free (cab);
    }

    return (NULL);
}

/*
 * Closes the cabinet 'cab', but leaves it on local storage. The cabinet can
 * be accessed later if opened with 'cabinet_open()'. The memory associated
 * with cab and it's folders is deallocated and should not be referenced
 * again.
 *
 *   Input:	cab	The cabinet to close.
 */
void
cabinet_close( cabinet cab )
{
    /*
     * if cab->name not set, this is a briefcase and should not be
     * deallocated with this function.
     */
    if ( cab )
	if ( cab->name ) {
	    free(cab->name);
	    free(cab);
	}
	else
	    fprintf(stderr, "cabinet_close(): Trying to close briefcase");
}

/*
 * Deletes the cabinet 'cab' and all folders associated with it. The memory
 * associated with cab and it's folders is deallocated and should not be
 * referenced again.
 *
 *   Input:	cab	The cabinet to close.
 */
void
cabinet_discard( cabinet cab )
{
    int		    cwdfd;
    DIR*	    dirp;
    char	    path[MAXPATHLEN];
    struct dirent*  dent;

    /* sanity checks */
    if ( cab && (cab->name == NULL) ) {
	fprintf(stderr, "cabinet_discard(): Trying to discard briefcase");
	return;
    }

    if ( cab == NULL )
	return;

    if (!*cabpath && !tac_init_cab_path()) return;

    strcpy(path,cabpath);
    strcat(path,cab->name);

    /* get current working dir, then change to cabinet */
    if ( (cwdfd = open(".", O_RDONLY)) == -1 ) {
	fprintf(stderr, "Could not get current working directory: %s",
		strerror(errno));
	return;
    }
    if ( chdir(path) == -1 ) {
	/* could not go to cabinet, so exit */
	fchdir(cwdfd);
	return;
    }

    /* delete folders and cabinet */
    if ( (dirp = opendir(path)) )
    {
	/* remove folders in cabinet */
	while ( (dent = readdir(dirp)) )
	    if ( strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..") )
		 if ( unlink(dent->d_name) == -1 )
		     fprintf(stderr, "Can't remove '%s': %s\n", dent->d_name,
			     strerror(errno));
	closedir(dirp);

	/* go back to working directory */
	fchdir(cwdfd);

	/* remove cabinet */
	if ( rmdir(path) == -1 )
	    fprintf(stderr, "Can't remove '%s': %s\n", path,
		    strerror(errno));
    }

    /* free allocated memory */
    free(cab->name);
    free(cab);
}

/*
 * Lists the folders of a local cabinet. The array of strings and the
 * strings themselves should be deallocated with 'free' after use. (Or
 * use the macro 'free_copy(sa)' to free it.)
 *
 *   Input:	cab	The cabinet to list folders from.
 *
 *   Returns:	A NULL-terminate array of strings that each represents the
 *		name of a folder in the cabinet, or NULL if error.
 */
string*
cabinet_list( cabinet cab )
{
    int		    count;
    DIR*	    dirp;
    char	    path[MAXPATHLEN];
    string*	    array;
    struct dirent*  dent;

    /* sanity checks */
    if ( cab && (cab->name == NULL) ) {
	fprintf(stderr, "cabinet_list(): Trying to list briefcase");
	return (NULL);
    }

    if ( cab == NULL )
	return (NULL);

    if (!*cabpath && !tac_init_cab_path()) return(NULL);
 
    strcpy(path, cabpath);
    strcat(path, cab->name);

    /* find all folders in cabinet */
    count = 0;
    array = NULL;
    if ( (dirp = opendir(path)) ) {
	while ( (dent = readdir(dirp)) )
	    if( strcmp(dent->d_name, ".") &&
		strcmp(dent->d_name, ".." ) &&
		strncmp(dent->d_name, "*#", 2) ) {
		if ( !array ) 
		    array = malloc( sizeof(string) * (count+2));
		else 
		    array = realloc(array, sizeof(string) * (count+2));
		if ( !array  )
		    return (NULL);
		if ( !(array[count++] = strdup(dent->d_name)) ) {
		    free_copy(array);
		    return (NULL);
		}
	    }
	closedir(dirp);
	if ( count )
	    array[count] = NULL;
    }

    return (array);
}

static int
cabinet_store( cabinet cab, string name, string* mdata )
{
    char     dpath[MAXPATHLEN];
    char     mpath[MAXPATHLEN];
    FILE*    fp_data;
    FILE*    fp_meta;
    string*  strings;

    if (!*cabpath && !tac_init_cab_path()) return(NULL);
 
    sprintf(dpath, "%s%s/%s", cabpath, cab->name, name);
    sprintf(mpath, "%s%s/*#%s", cabpath, cab->name, name);

    /* write data to file */
    if ( (fp_data = fopen(dpath, "w+")) )
    {
      if (!mdata[1]) {
	unlink(mpath);
	fwrite(mdata[0], strlen(mdata[0]), 1, fp_data);
      } else {
	if ( (fp_meta = fopen(mpath, "w+")) ) {
	  fputc('1', fp_meta);
	  fclose(fp_meta);

	  for ( strings=mdata; *strings; strings++ )
	    {
	      if ( **strings )
		fprintf(fp_data, "%s%s%s",
			strchr(*strings, ' ') || **strings == '{'
			? "{" : "",
			*strings,
			strchr(*strings, ' ') || **strings == '{'
			? "}" : "");
	      else
		fprintf(fp_data, "{}");
	      if ( strings[1] )
		fputc(' ', fp_data);
	    }
	}
	else {
	  fclose(fp_data);
	  return (-1);
	}
      }
      fclose(fp_data);
      return(0);
    }
    return (-1);
}

static string*
cabinet_fetch( cabinet cab, string name )
{
    int		 size;
    char	 dpath[MAXPATHLEN];
    char	 mpath[MAXPATHLEN];
    FILE*	 fp_data;
    string	 data;
    string*	 strings;
    struct stat	 afs;

    if (!*cabpath && !tac_init_cab_path()) return(NULL);

    sprintf(dpath, "%s%s/%s", cabpath, cab->name, name);
    sprintf(mpath, "%s%s/*#%s", cabpath, cab->name, name);

    strings = NULL;
    if ( (fp_data = fopen(dpath, "r")) )
    {
	/* find size of data on file */
	fseek(fp_data, 0, SEEK_END);
	size = ftell(fp_data);
	rewind(fp_data);

	/* allocate memory for data */
	if ( (data = malloc(size+1)) == NULL )
	    return (NULL);
	data[size] = '\0';

	/* read data from file */
	if ( fread(data, size, 1, fp_data) )
	    if ( (stat(mpath, &afs) == -1) ||
		 !(afs.st_mode & (S_IRUSR|S_IRGRP|S_IROTH)) ) {
		if ( (strings = calloc(2, sizeof(string))) )
		    *strings = data;
		else
		    free(data);
	    }
	    else {
		while ( isspace(data[strlen(data)-1]) )
		    data[strlen(data)-1] = '\0';
		strings = tac_str2array(data);
		free(data);
	    }
	else
	    return (NULL);
    }

    return (strings);
}

void
cabinet_delete( cabinet cab, string name )
{
    char	 dpath[MAXPATHLEN];
    char	 mpath[MAXPATHLEN];

    if (!*cabpath && !tac_init_cab_path()) return;

    sprintf(dpath, "%s%s/%s", cabpath, cab->name, name);
    sprintf(mpath, "%s%s/*#%s", cabpath, cab->name, name);

    if ( strcmp(name, ".") && strcmp(name, "..") )
      if ( unlink(dpath) == -1 )
	fprintf(stderr, "Can't remove '%s': %s\n", dpath,
		strerror(errno));
      else
	unlink(mpath);
}


/*****************************************************************************
 *
 *	Functions for creating and manipulating folders.
 *
 ****************************************************************************/
/*
 * Stores data into a folder in a briefcase/cabinet. It will either
 * overwrite old information if the folder already exists or create
 * a new folder with the supplied data. If no data is supplied, nothing
 * will happen.
 *
 *   Input:	bc	The briefcase/cabinet to fetch the folder from.
 *		name	The name of the folder.
 *		mdata	Multiple data as array of strings.
 *
 *   Returns:	0 on success, -1 if error.
 *
 *   Bugs:	strdup() calls are not checked for errors.
 */
int
folder_storeA( briefcase bc, string name, string* mdata )
{
    int	     hashidx;
    int	     count;
    folder   f;
    string*  str_p;
    string*  savedata;

    /* sanity checks */
    if ( !bc || !name || !*name || !mdata )
      return (-1);

    if ( bc->name )
      return (cabinet_store(bc, name, mdata));

    if ( !bc->f )
      return (-1);

    /* allocate array of strings with 'count' data elements */
    for ( count=0, str_p=mdata; *str_p; str_p++ )
      count++;
    if ( (savedata = malloc((count+1)*sizeof(string))) == NULL )
      return (-1);
    for ( str_p=savedata; *mdata; str_p++, mdata++ )
      *str_p = strdup(*mdata);
    *str_p = NULL;

    /* see if folder already exists */
    if ( (f = tac_find_folder(bc, name)) )
    {
	/* get rid of old data */
	free_copy(f->contents);
	/* set new contents in folder */
	f->contents = savedata;
    }
    else
    {
	/* allocate and initialize new folder */
	if ( (f = malloc(sizeof(folder_rec))) == NULL )
	{
	    /* error, so free data */
	    free_copy(savedata);
	    return (-1);
	}
	f->name = strdup(name);
	f->contents = savedata;

	/* link new folder into briefcase's hash buckets */
	hashidx = tac_hashit(name);
	f->next = bc->f[hashidx];
	f->prev = NULL;		/* as of now, prev is not used */
	bc->f[hashidx] = f;
    }

    return (0);
}

/*
 * Varargs stub for folder_storeA()
 */
int
folder_store( briefcase bc, string name, ... )
{
  int argc;
  int res;
  string* argv=NULL;
  string* str_p;
  va_list ap;

    /* count  arguments */
    va_start(ap, name);
    for ( argc=0; va_arg(ap, string); argc++ )
	;
    va_end(ap);
    if (argc>0) {
      /* read arguments */
      if ( (argv = malloc((argc+1)*sizeof(string))) == NULL )
	return (-1);
      va_start(ap, name);
      for ( str_p = argv; (*str_p = va_arg(ap, string)); str_p++ );
      va_end(ap);
      argv[argc] = NULL;
      res=folder_storeA(bc,name,argv);
      free(argv);
      return(res);
    }
    return(-1);
}


/*
 * Retrieves data from a folder in a briefcase. The returned array of
 * strings and the strings themselves should be deallocated with 'free'
 * after use. (Or use the macro 'free_copy(sa)' to free it.)
 *
 *   Input:	bc	The briefcase to fetch the folder from.
 *		name	The name of the folder.
 *
 *   Returns:	A pointer to a NULL-terminated array of strings that
 *		represents the contents of the folder, or NULL if the
 *		folder does not exist or an error ocurred.
 *
 *   Bugs:	calls to 'strdup()' are not checked for error
 */
string*
folder_fetch( briefcase bc, string name )
{
    int	     count = 0;
    folder   f;
    string*  tmp;
    string*  tmp2;
    string*  data;

    /* sanity checks */
    if ( !bc || !name || !*name )
	return (NULL);

    if ( bc->name )
	return (cabinet_fetch(bc, name));

    if ( !bc->f )
	return (NULL);

    if ( (f = tac_find_folder(bc, name)) )
    {
	for ( tmp=f->contents; *tmp; tmp++ )
	    count++;
	if ( (data=malloc((count+1)*sizeof(string))) == NULL )
	    return (NULL);
	for ( tmp=f->contents, tmp2=data; *tmp; tmp++, tmp2++ )
	    *tmp2 = strdup(*tmp);
	*tmp2 = NULL;

	return (data);
    }
    return (NULL);
}

/*
 * Looks at data in a folder from a briefcase. The value returned is a
 * pointer to memory used by the folder, and should NOT be deallocated
 * (or modified).
 *
 *   Input:	bc	The briefcase to fetch the folder from.
 *		name	The name of the folder.
 *
 *   Returns:	A pointer to a NULL-terminated array of strings that
 *		represents the contents of the folder, or NULL if the
 *		folder does not exist.
 */
string*	folder_look( briefcase bc, string name )
{
    folder  f;

    if ( (f = tac_find_folder(bc, name)) )
	return (f->contents);

    /* folder does not exist */
    return (NULL);
}

/*
 * Deletes a folder from a briefcase.
 *
 *   Input:	bc	The briefcase to delete the folder from.
 *		name	The name of the folder.
 */
void
folder_delete( briefcase bc, string name )
{
    int	    hashidx;
    folder  f, fpr;

    if ( bc->name ) {
      cabinet_delete(bc, name);
      return;
    }

    hashidx = tac_hashit(name);
    if ( !(f = bc->f[hashidx]) )
	return;

    /* find and unlink folder from briefcase hash buckets */
    if ( !strcmp(f->name, name) )
	bc->f[hashidx] = f->next;
    else
    {
	for ( fpr=f, f=f->next; f && strcmp(f->name, name); fpr=f, f=f->next )
	    ;
	if ( f )
	    fpr->next = f->next;
    }

    /* deallocate memory used by folder */
    tac_free_folder(f);
}

/*
 * Rotate the elements of a folder list left or right the specified amount.
 * Return the last element rotated. The returned string must be deallocated
 * after use.
 *
 *   Input:	bc	The briefcase that holds the folder.
 *		name	Name of the folder.
 *		dir	Direction of rotation: TAC_LEFT or TAC_RIGHT.
 *		amount	Rotate 'amount' elements.
 *
 *   Returns:	The last element to go from one end of the list to the other
 *		during the rotation, or NULL if error or amount is zero.
 */
string
folder_rotate( briefcase bc, string name, direction dir, int amount )
{
    int	     i, idx, count;
    folder   f;
    string   s1, s2;
    string*  str_p;

    /* find folder */
    if ( amount==0 || !(f = tac_find_folder(bc, name)) )
	return (NULL);

    /* count elements in folder data */
    for ( count=0, str_p=f->contents; *str_p; str_p++, count++ )
	;

    /* rotate elements 'amount' steps */
    amount *= dir;
    for ( i=0, idx=amount, s1=*f->contents; i<count; i++, idx+=amount )
    {
	if ( (idx %= count) < 0 )
	    idx += count;
	str_p = &f->contents[idx];
	s2 = *str_p;
	*str_p = s1;
	s1 = s2;
    }

    /* Find element to return */
    if ( amount < 0 )
    {
	/* rotated left, so return rightmost element */
	s1 = strdup(f->contents[count-1]);
    }
    else
    {
	/* rotated right, so return leftmost element */
	s1 = strdup(*f->contents);
    }

    return (s1);
}

/*
 * Shift the elements in a folder left or right the specified amount. The
 * elements shifted out are deleted, except the last shifted element, which
 * is returned. This element must be deallocated after use. Any optional
 * data (strings) will be added to the list from the opposite direction.
 * If the folder does not exist and there is no data, nothing will happen.
 * If the folder does not exist and there is data, a new folder is created.
 * If a folder becomes empty as a result of the shift it is silently removed.
 *
 *   Input:	bc	The briefcase that holds the folder.
 *		name	Name of the folder.
 *		dir	Direction of shift: TAC_LEFT or TAC_RIGHT.
 *		amount	Rotate 'amount' elements.
 *		mdata	Array of string pointers or NULL.
 *
 *   Returns:	The last element to be shifted out, or NULL if error, folder
 *		did not exist or amount is zero.
 *
 *   Bugs:	Does not check errors in 'strdup()' calls.
 */
string
folder_shiftA( briefcase bc, string name, direction dir, int amount, 
	       string* mdata )
{
    int	     argc, count, rest;
    folder   f;
    string   ret;
    string*  data;
    string*  str_p;
    string*  newdata;
    string*  argv;


    if (mdata) {
      for (argc=0,str_p=mdata; *str_p; str_p++, argc++);

      if ( (argv = malloc((argc+1)*sizeof(string))) == NULL )
	return (NULL);

      for (str_p=argv; *mdata; mdata++,str_p++)
	*str_p = strdup(*mdata);

      argv[argc]=NULL;
    } else {
      argc=0;
      argv=NULL;
    }

    /* change direction if amount is negative */
    if ( amount < 0 )
    {
	dir = (dir == TAC_LEFT) ? TAC_RIGHT : TAC_LEFT;
	amount = -amount;
    }

    /* if folder does not exist */
    if ( !(f = tac_find_folder(bc, name)) )
    {
      /* make new folder if optional data supplied */
      if ( argc ) {
	folder_storeA(bc, name, argv);
	free_copy(argv);
      }
      return (NULL);
    }

    /* count elements in folder; allocate memory for elements left after
     * shift and new data.
     */
    for ( count=0, str_p=f->contents; *str_p; str_p++ )
	count++;
    rest = ((count-amount) < 0) ? 0 : (count-amount);
    if ( (newdata = malloc((rest+argc+1)*sizeof(string))) == NULL )
    {
	free_copy(argv);
	return (NULL);
    }

    data = f->contents;
    if ( dir == TAC_RIGHT )
    {
	/* with right shift, data is added from the left */
	if ( argc )
	    memcpy(newdata, argv, argc*sizeof(string));
	/* if shift is less than folder contents, keep leftmost elements */
	if ( rest )
	    memcpy(&newdata[argc], data, rest*sizeof(string));
	/* figure out return value, and free shifted elements */
	if ( amount > count )
	{
	    ret = NULL;
	    free_copy(data);
	}
	else
	{
	  if (!amount) { 
	    ret=strdup(data[rest-1]); 
	  } else {
	    ret=data[rest];
	    for (str_p=&data[rest+1];*str_p;free(str_p++));
	  }
	  free(data);
	}
    }
    else
    {
	/* if shift is less than folder contents, keep rightmost elements */
	if ( rest )
	    memcpy(newdata, &data[amount], rest*sizeof(string));
	/* with left shift, data is added from the right */
	if ( argc )
	    memcpy(&newdata[rest], argv, argc*sizeof(string));
	/* figure out return value, and free shifted elements */
	if ( amount > count )
	{
	    ret = NULL;
	    free_copy(data);
	}
	else
	{
	  if (!amount) {
	    ret = strdup(data[0]); 
	  } else { 
	    ret = data[amount-1]; 
	    data[amount-1]=NULL;
	    for (str_p=data;*str_p;free(str_p++));
	  }
	  free(data);
	}
    }
    newdata[rest+argc] = NULL;
    if (argv) free(argv);

    /* delete folder if it contains no data */
    f->contents = newdata;
    if ( !f->contents[0] )
	folder_delete(bc, name);

    return (ret);
}

/*
 * Varargs stub for folder_shiftA()
 */
string
folder_shift( briefcase bc, string name, direction dir, int amount, ... )
{
  int argc;
  string* argv=NULL;
  string* str_p;
  string  res;
  va_list  ap;
  
  /* count optional arguments */
  va_start(ap, amount);
    for ( argc=0; va_arg(ap, string); argc++ )
	;
    va_end(ap);
    if (argc>0) {
      /* read optional arguments */
      if ( (argv = malloc((argc+1)*sizeof(string))) == NULL )
	return (NULL);
      va_start(ap, amount);
      for ( str_p = argv; (*str_p = va_arg(ap, string)); str_p++ );
      va_end(ap);
      argv[argc] = NULL;
    }
    res=folder_shiftA(bc,name,dir,amount,argv);
    free(argv);
    return(res);
}

/*
 *	Utility functions.
 */
static int
tac_hashit( string s )
{
    int val;

    for ( val=0; *s != '\0'; s++ )
	val += *s;

    return (val%TAC_BC_HASHSIZE);
}

static folder
tac_find_folder( briefcase bc, string name )
{
    int	    hashidx;
    folder  f;

    hashidx = tac_hashit(name);
    for ( f=bc->f[hashidx]; f && strcmp(f->name, name); f=f->next )
	;

    return (f);
}

static string*
tac_str2array( string str_p )
{
    int	     count = 0;
    int      size;
    int	     leftbraces;
    string   str;
    string   s_ptr;
    string   arc=str_p;
    string   end;
    string*  arr;

    for ( str=arc, arr=NULL, leftbraces=0; *arc; str++ )
	switch ( isspace(*str) ? ' ' : *str )
	    {
	    case '{':
		leftbraces++;
		break;
	    case '}':
		leftbraces--;
		break;
	    case '\\':
		str++;
		break;
	    case ' ':
	    case '\n':
		if ( leftbraces )
		    break;
		if ( str == arc ) {
		    arc++;
		    break;
		}
	    case '\0': /* or ' ' above */
		if ( *arc == '{' )
		    size = str-(arc++)-1;
		else
		    size = str-arc+1;
		if ( !arr ) 
		    arr = malloc( (count+2)*sizeof(string));
		else
		    arr = realloc(arr, (count+2)*sizeof(string));
		arr[count] = s_ptr = malloc((size_t) size--); 
		for (; size>0; size--) {
		    if (*arc=='\\') {
			switch (arc[1]) {
			case '\\':
			case '{':
			case '}':
			    size--; arc++; break;
			default:
			    break;
			}
		    }
		    *s_ptr++ = *arc++; 
		}
	        *s_ptr='\0';
		
		count++;
		if ( *str == '\0' )
		    arc = str;
		else 
		    arc = str+1;
		break;
	    default:
		break;
	    }
		
		arr[count] = NULL;
    return (arr);
}

int tac_init_cab_path(void) 
{
  int t=0;
  string path;

  if( !(path = getenv(TAC_ENV_PATH)) ) {
    fprintf(stderr, "%s not configured", TAC_ENV_PATH);
    return (0);
  }
  sprintf(cabpath, "%s%s" TAC_CABINETS, path,
	  path[strlen(path)-1] == '/' ? "" : "/");
#ifdef NFS_HACK
  t=strlen(cabpath);
  gethostname(&cabpath[t],MAXPATHLEN-t-1);
  if (path=strchr(&cabpath[t],'.')) *path='\0';
  strcat(cabpath,"/");
#endif
  return(1);
}
