.TH ag_elk 3 "TACOMA 1.1.3" "Agents"

.SH NAME
ag_elk - agent which executes Scheme code.
.SH DESCRIPTION

The ag_elk agent evaluates all Scheme code found in the folder
SCMCODE.  Ag_elk then rotates the SCMEVAL folder, and evaluates the
first element found there.  This makes it possible to have a Scheme
program defining some functions in the SCMCODE folder, and call some
of these functions from the SCMEVAL folder.  When the client agent
want to utilize some new functions (or call some old functions with
different arguments), it may simply store some new values into the
SCMEVAL folder and leave the SCMCODE folder untouched.
.SS Folder inputs

.TP
SCMCODE
A folder containing elements with Scheme code to be evaluated upon
startup of ag_elk.  The elements are evaluated in the order of which
they appear in the folder.
.TP
SCMEVAL
A folder containing Scheme code to be evaluated after the evalutation
of the SCMCODE folder.  However, only the first element of this folder
is evaluated, and the folder is then rotated to the left.
.P

.SH THE TACOMA SCHEME INTERFACE

With ag_elk installed, it is possible to write TACOMA agents using
Elk.  To do this, you have to include the line

    (require 'tacoma)

in your Scheme code.  This enables you to make use of the following
procedures, and thereby providing you with a Scheme interface to the
TACOMA primitives.

The usual naming conventions for Scheme procedures are used.  Ie.
procedures ending with \`?' are predicates, and allways return a
booelan value.  Procedures that store values into previous allocated
locations end in \`!', and procedures that convert between object
types containg the string \`->'.
.br
.SS Procedures for operations on briefcases.

.TP
(briefcase? \fIobj\fP)
Return #t if \fIobj\fP is of type breifcase, #f otherwise.
.TP
(bc-create)
Create a briefcase and return it.
.TP
(bc-discard! \fIbc\fP)
Deletes the briefcase, \fIbc\fP, and all folders associated with it.
Returns the value #v.
.TP
(bc-list \fIbc\fP)
Returns a list of all folders within briefcase, \fIbc\fP.
.TP
(archive \fIbc\fP) or (briefcase->string \fIbc\fP)
Archives briefcase, \fIbc\fP, into a string and returns the string.
.TP
(unarchive! \fIbc archive\fP)
Unarchives the archive into briefcase, \fIbc\fP, and returns
the briefcase.
.br
.SS Procedures for operations on cabinets.

.TP
(cabinet? \fIobj\fP)
Return #t if \fIobj\fP is of type cabinet, #f otherwise.
.TP
(cabinet-open \fIname\fP [\fIcreate\fP])
Open local cabinet, \fIname\fP, and return it.  If a second argument
\fB\'create\fP is specified, create the cabinet if it does not exist.  If an
error occured, or cabinet did not exist (and was not created), return #f.
.TP
(cabinet-close! \fIcab\fP)
Closes cabinet, \fIcab\fP, but leaves it on local storage so that it
may be opened later.
.TP
(cabinet-discard! \fIcab\fP)
Deletes cabinet, \fIcab\fP, and all folders associated with it.
.TP
(cabinet-list \fIcab\fP)
Returns a list folders of a local cabinet, \fIcab\fP, or #f if error.
.br
.SS Procedures for operations on folders.

.TP
(folder-store! \fIbc name data\fP [\fIdata\fP ...])
Store data within folder, \fIname\fP, located in briefcase, \fIbc\fP.
Returns #t on success, #f if error.
.TP
(folder-fetch \fIbc name\fP)
Return contents of folder, \fIname\fP, located in briefcase, \fIbc\fP, as
a list of strings.
.TP
(folder-delete! \fIbc name\fP)
Delete contents of folder, \fIname\fP, located in briefcase, \fIbc\fP.
Returns #v.
.TP
(folder-shift! \fIbc name dir amount\fP [\fIdata\fP ...])
Shift folder, \fIname\fP, located in briefcase, \fIbc\fP, in
direction, \fIdir\fP (left or right) by amount, \fIamount\fP.  If
\fIdata\fP is supplied, shift it into the other end of the folder.
Returns the last element shifted out of the folder, or the leftmost
(or rightmost) element in folder if no shifting was done.
.TP
(folder-rotate! \fIbc name dir amount\fP)
Rotate folder, \fIname\fP, located in briefcase, \fIbc\fP, in
direction, \fIdir\fP (left or right) by amount, \fIamount\fP.  Returns
the last element rotated.
.TP
(folder-pop! \fIbc name\fP)
Macro for deleting and returning the leftmost element of folder
\fIname\fP in briefcase, \fIbc\fP.
.TP
(folder-add! \fIbc name data\fP)
Macro for adding an element to the left of folder, \fIname\fP, in
briefcase, \fIbc\fP.
.TP
(folder-push! \fIbc name\fP)
Macro for adding an element to the right of folder, \fIname\fP, in
briefcase, \fIbc\fP.
.TP
(folder-top \fIbc name\fP)
Macro for looking at the rightmost element of folder, \fIname\fP, in
briefcase, \fIbc\fP, without deleting it.
.br
.SS The meet functionality.

.TP
(meet \fIagent\fP \fIbc\fP [\fIflags\fP])
Meet agent, \fIagent\fP, with briefcase, \fIbc\fP.  Flags that may be
specified include: \fBsync (rpc), local, nomerge, ignerr, crypt, sign,
path\fP and \fBnolog\fP.  They are specified as a list of symbols.  Eg:

    (meet "ag_elk" bc '(sync nomerge))

Upon successful completion, \fBmeet\fP returns #t.  Otherwise, #f is
returned, and \fBmeet-error\fP is set to indicate the error.  The
values that \fBmeet-error\fP may take, include \fBOK, notreached,
refused, nosite, ag-dead, ag-notfound\fP and \fBinternal\fP (see
\fImeet(3C)\fP).  A negative integer may also be returned to indicate
an error returned by the local agent.
.TP
(meet-strerror [\fIerrorcode\fP])
Compound procedure that returns a descriptive string of the indicated
\fIerrorcode\fP.  If no errorcode is specified, describe the errorcode
indicated by \fBmeet-error\fP.
.br
.SS Notes
All procedures that that expect a string as an argument (eg. folder
name), may also receive a symbol.  The symbolname is then used as a
string.

To help the programmer putting Scheme code into folders, a macro
(string arg) is defined, that converts a given argument to a
string.   The following code will then store a scheme expression
into the SCMCODE folder.

  (folder-store! bc 'SCMCODE (string '(set! autoload-notify #f)))

A Scheme programmer usually does not free objects when he is done with
them.  He leaves that up to the garbage collector.  Procedures like
(bc-discard ...) therefore has little meaning in a Scheme
implementation.  They are included here however, to conform with the
TACOMA primitives used in other programming languages.  The programmer
is of course free to ignore them if he want to.  The garbage collector
will clean things up anyway.
.br
.SH INVOCATION OF AG_ELK
When ag_elk evaluates the incoming Scheme code, the evaluated code may
access the incoming briefcase (and all its folders) by referencing the
briefcase \fBinc-bc\fP.  Eg.

    (folder-fetch inc-bc 'RETURNHOST)

will return the contents of the RETURNHOST folder supplied with the
incoming briefcase.

The code specified in SCMCODE or SCMEVAL may also use the TACOMA
procedures specified above without doing (require 'tacoma).  This,
however, only applies to primitive procedures (ie. not macros or
compound procedures).
.br
.SH FILES
$TACOMAPATH/sysagents/ag_elk
.br
/usr/local/elk/runtime/obj/tacoma.o
.br
/usr/local/elk/runtime/scm/tacoma.scm
.br
.SH SEE ALSO
elk(1), TACOMA(3)
.br
.SH AUTHOR

Ag_elk was written by Espen Skoglund, University of Tromsoe
<espensk@stud.cs.uit.no>.
