package Tacoma;

use strict;
use vars qw($VERSION @ISA @EXPORT);

require Exporter;
require DynaLoader;

@ISA = qw(Exporter DynaLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw(
	meet
);
$VERSION = '1.2';

bootstrap Tacoma $VERSION;

# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is the documentation for the module.

=head1 NAME

Tacoma - Perl interface to the I<Tacoma(3)> system.

=head1 SYNOPSIS

C<use Tacoma;>

=head1 DESCRIPTION

With support for Perl and B<ag_perl> installed, it is possible to write
TACOMA agents using Perl.  To do this, you have to include the line

C<    use Tacoma;>

in your Perl code.  This enables you to make use of the following
subs, which provides you with a Perl interface to the TACOMA primitives.

=head2 Subs operating on briefcases.

These are the primitives that operate on briefcases:

=over 4

=item I<$bc> = new Briefcase;

Create a briefcase and return it.

=item I<@folders> = I<$bc>->list;

Returns a list of all folders within briefcase I<$bc>, or empty list if error.

=item I<$arc> = I<$bc>->archive;

Archives the briefcase I<$bc> into a string and returns the string.

=item I<$abc> = I<$bc>->unarchive(I<$arc>);

Unarchives the archive I<$arc> into briefcase I<$bc> and returns the
briefcase.

=back

=head2 Subs operating on cabinets.

These are the primitives that operate on cabinets:

=over 4

=item I<$cab> = new Cabinet(I<$name> [, C<'create'>]);

Open local cabinet I<$name> and return it.  If the second optional
argument C<'create'> is specified, create the cabinet if it does not exist.
Returns undefined if an error occurred, or cabinet did not exist (and was
not created).

=item I<$cab>->discard;

Deletes cabinet I<$cab> and all folders associated with it.

=item I<@folders> = I<$cab>->list;

Returns a list of all folders within cabinet I<$cab>, or empty list if error.

=back

=head2 Subs operating on folders.

These are the primitives that operate on folders.  The primitives
C<store>, C<fetch> and C<delete> also work with cabinets.

=over 4

=item I<$status> = I<$bc>->store(I<$name>, I<@data>);

Store list of data I<@data> within folder I<$name> located in briefcase I<$bc>.
Returns 0 on success and -1 on failure.

=item I<@data> = I<$bc>->fetch($name);

Return contents of folder I<$name> located in briefcase I<$bc> as a list of
strings.

=item I<$bc>->delete(I<$name>);

Deletes folder I<$name> located in briefcase I<$bc>.

=back

The rest of these primitives just work with briefcases:

=over 4

=item I<$data> = I<$bc>->shift(I<$name>, I<$dir>, I<$amount>, I<@data>);

Shift folder I<$name>, located in briefcase I<$bc>, in direction I<$dir>
(C<'left'> or C<'right'>) by I<$amount>.  If I<@data> is supplied, shift it
into the other end of the folder.  Returns the last element shifted out of
the folder, or the leftmost (or rightmost) element in folder if no shifting
was done.

=item I<$data> = I<$bc>->rotate(I<$name>, I<$dir>, I<$amount>);

Rotate folder I<$name>, located in briefcase I<$bc>, in direction I<$dir>
(C<'left'> or C<'right'>) by I<$amount>.  Returns the last element rotated.

=item I<$data> = I<$bc>->pop(I<$name>);

Deletes and returns the leftmost element of the folder I<$name> in the
briefcase I<$bc>.

=item I<$bc>->push(I<$name>, I<$data>);

Adds I<$data> to the right of the folder I<$name> in the briefcase I<$bc>.

=item I<$data> = I<$bc>->top(I<$name>);

Looks at the rightmost element of folder I<$name> in briefcase I<$bc>
without deleting it.

=item I<$bc>->add(I<$name>, I<$data>);

Adds an element to the left of a folder I<$name> in briefcase I$<bc>.

=back

=head2 The meet functionality.

=over 4

=item I<$status> = meet(I<$agent>, I<$bc> [, I<@flags>]);

Meet I<$agent> with briefcase I<$bc>.  The optional list of flags which may
be specified include: C<'sync'> (C<'rpc'>), C<'local'>, C<'nomerge'>, C<'ignerr'>, C<'crypt'>, C<'sign'>, C<'path'> and C<'nolog'>.
They are specified as a list of symbols, eg:

C<meet('ag_perl', $bc, 'sync', 'nomerge');>

Upon successful completion, B<meet> returns 0.  Otherwise, an error-code is
returned (see L<meet(3C)>).  A negative integer may also be returned
indicating an error returned by the local agent.

=back

=head2 Notes

The TACOMA primitives dealing with free'ing resources (B<bc_discard>, etc)
are not visible to the Perl programmer.  This Perl extension automagically
deallocates any resources used by briefcases, cabinets, etc., when the
variables are no longer in use.

=head1 AUTHOR

Magne Rasmussen, University of Tromsoe <magner@stud.cs.uit.no>.

=head1 SEE ALSO

perl(1).
ag_perl(1).
TACOMA(3).

=cut
