.TH ag_tcl 1 "TACOMA 1.1" "Agents"
.SH NAME
ag_tcl \- agent which executes Tcl code
.sp
ag_wish \- agent which executes Tcl/Tk code
.SH DESCRIPTION
The ag_tcl agent evaluates the Tcl code found in all the folders
named in the TCLCODE folder list. This folder \fBmust\fR
exist and the list should be terminated with an empty item.
Ag_tcl then rotates TCLMAIN to determine which procedure to call. If TCLMAIN
is empty or non-existent, a Tcl procedure called 'main' will
be called by default. This procedure is called with the supplied
briefcase as it's first and only argument.
.SS Note
It is quite legal to specify a non-existing folder in the
TCLCODE list, but at least one folder must exist. This
is the folder containing the main procedure to be called. This
behavior allows the application programmer to drop (delete) code
folders as the agent travels in order to reduce the size of the
\fIbriefcase\fR.
.SS Folder inputs
.TP 
TCLCODE 
a folder containing a list of folders each containing
Tcl code to be evaluated before calling the main
procedure.
.TP 
TCLMAIN
a list of procedures. The first procedure named on this
list will be called at this site, and since this folder
is rotated, it could contain a predefined list of
procedures to call at different sites. However if the
same procedure should be called at each site, this
folder should contain only the name of that procedure.
.SH ""
The ag_wish agent functions exactly like ag_tc,l but it is also able
to cope with the Tk extension to Tcl. Most likely, your main routine
will simply set up the necessary windows, buttons and the actions to be
performed. So your main routine should return zero (success) if this
succeeds. If you do not return a zero, ag_wish will terminate
immediately. However, to terminate the agent based on Tk events,
your code should set the \fBag_wish_exit\fR variable, in order to tell
the underlying TACOMA system that your Tk code has finished. For
example, this is a button which will terminate the agent (at the
request of the user at the site the agent is currently executing):

.RS
button .c -text "Quit" -command "set ag_wish_exit 0"
.RE

The value \fBag_wish_exit\fR is set to becomes the result code of the
agent. Setting it to zero indicates a successful termination, while
setting it to a negative value indicates failure. Further
information should be supplied in the STATUS folder.
.SS Important
Since Tcl/Tk evaluates scripts given to the \fB-command\fR option in
the context
of the lowest stack level, references to the briefcase of an agent must
always be made to \fIloc_bc\fR, regardless of the current local binding. For example:

.RS
button .c -text "Cancel" -command "folder_store \fIloc_bc\fR RESULT 0"
.RE
.SH ""
The needed folders are also
named differently:
.SS Folder inputs
.AS briefcase WISHCODE
.AP briefcase WISHCODE in
a folder containing a list of folders each containing
Tcl/Tk code to be evaluated before calling the main
procedure.
.AP briefcase WISHMAIN in/out
a list of procedures. The first procedure named on this
list will be called at this site. Since this folder
is rotated, it could contain a predefined list of
procedures to call at different sites. However, if the
same procedure should be called at each site, this
folder should contain only the name of that procedure.
.AP global ag_wish_exit out
TACOMA waits for this variable to be modified before terminating the agent, and sets the result code of the agent accordingly.

.SH SEE ALSO
meet(n), folder_rotate(n)

.SH AUTHOR
Nils Peter Sudmann, University of Tromsoe (nilss@stud.cs.uit.no)
Concept by Dag Johansen (dag@cs.uit.no), Robbert van Renesse
(rvr@cs.cornell.edu) and Fred B. Schneider (fbs@cs.cornell.edu).
